from square_auth.client_credentials import ClientCredentials
client_credentials = ClientCredentials()

from square_skill_helpers.square_api import SquareAPI
from square_skill_helpers.data_api import DataAPI
# from square_skill_helpers.model_api_deprecated import ModelAPI
from square_skill_helpers.model_api import ModelAPI
